@@ -43,6 +43,17 @@
 #define SPG_YIELD_KMV_HASH_GROUPS 12
 #define SPG_YIELD_MKMV_HASH_GROUPS 13
 
+/* Whether the data objects are structs instead of just pointers */
+static int unwrap_structs;
+
+/* External functions defined by ruby-pg when data objects are structs */
+PGconn* pg_get_pgconn(VALUE);
+PGresult* pgresult_get(VALUE);
+
+/* Normalize access to data objects for both old and new versions of pg gem */
+#define GetPGconn(_val, _var) if (unwrap_structs) {Check_Type(_val, T_DATA); _var = pg_get_pgconn(_val);} else {Data_Get_Struct(_val, PGconn, _var);}
+#define GetPGresult(_val, _var) if (unwrap_structs) {Check_Type(_val, T_DATA); _var = pgresult_get(_val);} else {Data_Get_Struct(_val, PGresult, _var);}
+
 static VALUE spg_Sequel;
 static VALUE spg_Blob;
 static VALUE spg_BigDecimal;
@@ -235,6 +246,27 @@ static VALUE parse_pg_array(VALUE self, VALUE pg_array_string, VALUE converter)
   char *word = RSTRING_PTR(buf);
   int index = 1;
 
+  if (array_string_length == 0) {
+    rb_raise(rb_eArgError, "unexpected PostgreSQL array format, empty");
+  }
+
+  switch (c_pg_array_string[0]) {
+    case '[':
+      /* Skip explicit subscripts, scanning until opening array */
+      for(;index < array_string_length && c_pg_array_string[index] != '{'; ++index)
+        /* nothing */;
+
+      if (index >= array_string_length) {
+        rb_raise(rb_eArgError, "unexpected PostgreSQL array format, no {");
+      } else {
+        ++index;
+      }
+    case '{':
+      break;
+    default:
+      rb_raise(rb_eArgError, "unexpected PostgreSQL array format, doesn't start with { or [");
+  }
+
   return read_array(&index, c_pg_array_string, array_string_length, word, converter
 #ifdef SPG_ENCODING
 , enc_get_index(pg_array_string)
@@ -568,7 +600,6 @@ static void spg_set_column_info(VALUE self, PGresult *res, VALUE *colsyms, VALUE
       case 17:
       case 20:
       case 21:
-      case 22:
       case 23:
       case 26:
       case 700:
@@ -609,16 +640,20 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
   VALUE pg_value;
   char type = SPG_YIELD_NORMAL;
 
+  if (!RTEST(rres)) {
+    return self;
+  }
+  GetPGresult(rres, res);
+
 #ifdef SPG_ENCODING
   int enc_index;
   enc_index = enc_get_index(rres);
 #endif
 
-  Data_Get_Struct(rres, PGresult, res);
   ntuples = PQntuples(res);
   nfields = PQnfields(res);
   if (nfields > SPG_MAX_FIELDS) {
-    rb_raise(rb_eRangeError, "more than %d columns in query", SPG_MAX_FIELDS);
+    rb_raise(rb_eRangeError, "more than %d columns in query (%ld columns detected)", SPG_MAX_FIELDS, nfields);
   }
 
   spg_set_column_info(self, res, colsyms, colconvert);
@@ -851,7 +886,7 @@ static VALUE spg_supports_streaming_p(VALUE self) {
 #if HAVE_PQSETSINGLEROWMODE
 static VALUE spg_set_single_row_mode(VALUE self) {
   PGconn *conn;
-  Data_Get_Struct(self, PGconn, conn);
+  GetPGconn(self, conn);
   if (PQsetSingleRowMode(conn) != 1) {
       rb_raise(spg_PGError, "cannot set single row mode");
   }
@@ -875,11 +910,11 @@ static VALUE spg__yield_each_row(VALUE self) {
 
   rconn = rb_ary_entry(self, 1);
   self = rb_ary_entry(self, 0);
-  Data_Get_Struct(rconn, PGconn, conn);
+  GetPGconn(rconn, conn);
 
   rres = rb_funcall(rconn, spg_id_get_result, 0);
   rb_funcall(rres, spg_id_check, 0);
-  Data_Get_Struct(rres, PGresult, res);
+  GetPGresult(rres, res);
 
 #ifdef SPG_ENCODING
   int enc_index;
@@ -927,7 +962,7 @@ static VALUE spg__yield_each_row(VALUE self) {
 
     rres = rb_funcall(rconn, spg_id_get_result, 0);
     rb_funcall(rres, spg_id_check, 0);
-    Data_Get_Struct(rres, PGresult, res);
+    GetPGresult(rres, res);
   }
   rb_funcall(rres, spg_id_clear, 0);
 
@@ -938,7 +973,7 @@ static VALUE spg__flush_results(VALUE rconn) {
   PGconn *conn;
   PGresult *res;
   VALUE error = 0;
-  Data_Get_Struct(rconn, PGconn, conn);
+  GetPGconn(rconn, conn);
 
   while ((res = PQgetResult(conn)) != NULL) {
     if (!error) {
@@ -1043,6 +1078,9 @@ void Init_sequel_pg(void) {
   if (rb_eval_string("Date.today.instance_variable_get(:@ajd)") != Qnil) {
     spg_id_Rational = rb_intern("Rational");
   }
+  if (rb_eval_string("defined?(PG::TypeMapAllStrings)") != Qnil) {
+    unwrap_structs = 1;
+  }
 
   c = rb_funcall(spg_Postgres, cg, 1, rb_str_new2("Dataset"));
   rb_define_private_method(c, "yield_hash_rows", spg_yield_hash_rows, 2);
